  (*

    NPRG: N-Precedence Relation Grammars

    A n-precedence relation grammar is a set of allowable n-precedence relations.

    Requirements: The Tools and FSA libraries must be loaded.

  *)



(*
  SETTING UP PARAMETERS
*)

let usage () = print_endline (
    "\nBASIC USAGE: prg [subcommand] [infile]\n\n"^
      "Subcommands are:\n\n"^
      "find\t\t"^
      "accepts\t\t"^
      "2dot\n\n"^
      "See the documentation for more information.\n")


let cmd_list = List.tl (Array.to_list Sys.argv)
  
let error = ref false
let wd = ref ""
let num = ref 2
let wb = ref "#"
let f1 = ref ""
let f2 = ref ""

let rec process_args list = 
  match list with
    | [] -> ()

    | "-d"::(seg_delim::t) -> wd := seg_delim; process_args t

    | "-b"::(wd_bound::t) -> wb := wd_bound; process_args t

    | "-d"::[]
    | "-b"::[]
    | "-n"::[]  -> error := true

    | arg::t ->
	match (!f1,!f2) with
	  | ("","") -> f1 := arg; process_args t
	  | (_,"") -> f2 := arg; process_args t
	  | _ -> error := true;;

(* note that f2 is a nonempty string iff f1 is a nonempty string *)

process_args (List.tl cmd_list);;

(* print_endline ("Err:"^(string_of_bool !error));; *)
(* print_endline ("Wd:"^(!wd));; *)
(* print_endline ("F1:"^(!f1));; *)
(* print_endline ("F2:"^(!f2));; *)

module WD =
struct
  let delim = !wd 
  let lb = ""
  let rb = ""
end

(* Vocab is a module in FSA library.*)
module PR = Xlist.Make(Delim.Comma_prs)(Vocab)
module VSet = Xset.Make(Delim.Comma_cbs)(Vocab)
module PRSet = Xset.Make(Delim.Semi_cbs)(PR)
(* module Word = Xlist.Make(Delim.None)(Vocab) *)
module Word = Xlist.Make(WD)(Vocab)
module WordSet = Xset.Make(Delim.Newline_no_bs)(Word)



let wboundary = Vocab.of_string !wb

(* pads word boundary symbols onto the beginning of words *)
let padwb w = 
  Word.build wboundary (Word.rev (Word.build wboundary (Word.rev w)))

(* 
   Adds symbol to the front of every 
*)
let rec add_h_to_list symbol word =
  Word.fold_left
    (fun prs v -> PRSet.add (PR.build symbol (PR.build v PR.empty)) prs)
    PRSet.empty word

(* finds the precedence relations in a word. *)
let rec of_word w = 
  let len = Word.length w in
  if len < 2 then PRSet.empty
  else
    let h = Word.hd w and t = Word.tl w in
    PRSet.union (add_h_to_list h t) (of_word t)
    

(* finds the precedence relations in a set of words. *)
let prs_of_wordset ws = 
  WordSet.fold
    (fun w prs -> PRSet.union (of_word (padwb w)) prs)
    ws PRSet.empty
    

let prs_of_wordfile filename =
  (* each word should be on its own line *)
  let gprs = ref PRSet.empty in
  let f s = 
    let w = Word.of_string s in
    gprs := (PRSet.union !gprs (of_word (padwb w)))
  in
  File.iter filename f; !gprs

let get_sigma prs = 
  let add v sigma = 
    if Vocab.compare v wboundary = 0 then sigma 
    else VSet.add v sigma
  in
  PRSet.fold
    (fun pr vs ->
      let a = PR.hd pr and b = PR.hd (PR.tl pr) in
      add a (add b vs)
    )
    prs
    VSet.empty

let print_g g = 
  PRSet.iter (fun pr -> PR.print pr) g

(* 
   returns true iff g accepts w. I.e. if the precedence relations in w are a
   subset of g.
*)
let accepts g w = PRSet.subset (of_word w) g

module Symbol = Symbols.Make(Vocab)
module NodeSet = Xset.Make(Delim.Point_abs)(Node)
module NodeSetSet = Xset.Make(Delim.Newline_no_bs)(NodeSet)
module Edge = Edges.Make(Node)(Symbol)(Delim.Colon_no_bs)
module EdgeSet = Xset.Make(Delim.Newline_no_bs)(Edge)

(* Now we can make the FSA module. *)
module FSA = Acceptors.Make(Vocab)(Word)(WordSet)(Symbol)
  (Node)(NodeSet)(NodeSetSet)(Edge)(EdgeSet)(Delim.Excl_no_bs);;


let node_of_vset vset = Node.of_string (VSet.to_string vset)
let vset_of_node node = VSet.of_string (Node.to_string node)

exception Not_in_grammar

let terminus_node node a g =
  let vs = vset_of_node node
  in
  let for_all_v_in_vs_va_in_g = 
    VSet.fold
      (fun v bool -> 
	let pr = PR.build v (PR.build a PR.empty) in
	bool && PRSet.mem pr g 
      ) vs true
  in
  if for_all_v_in_vs_va_in_g then node_of_vset (VSet.add a vs)
  else raise Not_in_grammar

  (* if for all v in vs, (v,a) is in g then VSet.add a vs 
     else raise Not_in_grammar *)

let to_fsa g sigma =
  let start = node_of_vset (VSet.singleton wboundary)
  in
  let rec make_delta nodes_to_do finished_nodes edgeset =
    if NodeSet.subset nodes_to_do finished_nodes then edgeset
    else
      let new_edges,new_nodes_to_do =
	NodeSet.fold
	  (fun node (es1,ns1) ->
	    let new_edges2,new_nodes2 = 
	      VSet.fold
		(fun v (es2,ns2) ->
		  try
		    let new_node = 
		      terminus_node node v g
		    in
		    let new_edge = 
		      Edge.make node new_node (Symbol.of_x v)
		    in
		    (EdgeSet.add new_edge es2,
		    NodeSet.add new_node ns2)
		  with Not_in_grammar -> (es2,ns2)
		) 
		sigma (es1,ns1)
	    in
	    (EdgeSet.union es1 new_edges2, NodeSet.union ns1 new_nodes2)
	  ) 
	  nodes_to_do (EdgeSet.empty,NodeSet.empty)
      in
      make_delta 
	new_nodes_to_do 
	(NodeSet.union finished_nodes nodes_to_do)
	(EdgeSet.union new_edges edgeset)
  in
  let edges = make_delta (NodeSet.singleton start) 
    NodeSet.empty EdgeSet.empty
  in
  let nodes = (FSA.nodes (FSA.make NodeSet.empty NodeSet.empty edges)) (* all nodes *)
  in
  let finals = 
    NodeSet.fold
      (fun n ns -> try 
	  let _ = terminus_node n wboundary g in
	  NodeSet.add n ns
	with Not_in_grammar -> ns
      ) nodes NodeSet.empty
  in
  FSA.make 
    (NodeSet.singleton start)
    finals
    edges


(* MAIN PROGRAM BEGINS HERE *)

let raise_error () = 
  let () = List.iter (fun x -> print_string (x^":")) cmd_list in
  let () = print_endline "" in
  let () = print_endline "Unrecognized arguments" in 
  usage ();;

let main () =
  let subcommand = 
    try List.hd cmd_list 
    with _ -> print_endline "e1"; raise_error (); failwith ""
  in
  if !error then begin print_endline "e2"; raise_error () end
  else
    match subcommand with
      | "accepts" ->
	  (*
	    Write this so that -g is a flag needed to indicate the grammar. If -g is
	    followed by nothing then the grammar is entered from standard input,
	    otherwise its argument is a filename.
	  *)
	  begin
	    let g,w = 
	      match (!f1,!f2) with
		  (*		| ("","")   (PRSet.of_string (File.read ()), input_line stdin) *)
		| (_,"") -> raise_error (); failwith "" (*(PRSet.of_file a, input_line stdin)*)
		| (a,b) -> (PRSet.of_file a, b)
	    in
	    print_endline (string_of_bool (accepts g (Word.of_string w)))
	  end

      | "2fsa" -> 
	  begin
	    let g = 
	      match (!f1,!f2) with
		| ("","") -> File.of_string "temp_pr_grmr" (File.read ());
		    PRSet.of_file "temp_pr_grmr"
		| (a,"") -> PRSet.of_file a
		| _ -> raise_error (); failwith ""
	    in
	    FSA.print (to_fsa g (get_sigma g))
	  end
	    
      | "find" ->
	  begin
	    match (!f1,!f2) with
	      | ("","") -> print_g (prs_of_wordset (WordSet.of_string (File.read ())))
	      | (a,"") -> print_g (prs_of_wordfile a)
	      | _ -> print_endline "e3"; raise_error ()
	  end

      | _ -> print_endline "e4"; raise_error ();;


main ();;
